with Ada.Characters.Handling,
     Ada.Unchecked_Deallocation,
     Http_Urls;

package body Html_Scan is
  use Ada.Characters.Handling;

  Note_Found_Menu_Args_Link: constant Boolean := False;

  procedure Get_Links(Me      : in     Character; -- ID during debuggin
                      Debug_Flag: in   Boolean;
                      Source_Name: in  String) is
    Extended_Line: String(1 .. Max_Line_Length);
    Left_Extension: constant Natural := 7; -- allows s-7
    Right_Extension: constant Natural := 15; -- allows si+15
    Line : String
      renames Extended_Line(Left_Extension + 1 .. Max_Line_Length - Right_Extension);
    Last : Natural;
    Si,
    Wi  : Natural;
    type Scanning_States is (In_Content, In_Content_String, In_Bracket,
      In_Bracket_Word, In_Bracket_String, Expect_Eq, After_Eq, In_Url,
      In_Unquoted_Url, Skip_To_End_Bracket, In_Comment,
      In_Script, In_Script_Menu, In_Script_Menu_Args, In_Script_Menu_Args_Qs,
      In_Script_Menu_Args_Link);
    -- Note: In_Script handles javascript menus in a very ad hoc way!
    State   : Scanning_States := In_Content;
    Target  : String(1 .. 900); -- scanned urls
    Di      : Natural;
    Meta_Seen,
    Http_Equiv_Seen: Boolean;
    Any_Log_Output: Boolean := False;
    Handling_Base: Boolean := False;
    type Base_Ptr is access String;

    procedure Free is new Ada.Unchecked_Deallocation(String, Base_Ptr);

    Current_Base: Base_Ptr := new String'(Source_Name);

    procedure Note(S : in String) is
    begin
      -- Note: The caller is adding "Me", so we don't have to do it.
      if not Any_Log_Output then
        Any_Log_Output := True;
        Log("Scan:" & Source_Name);
      end if;
      Log(S);
    end Note;

  begin
    State := In_Content;
    Last := 0;
    Si := Line'First;
    loop
      if Si > Last then
        Get_Line(Line, Last);
        exit when Last < Line'First; -- done
        -- drop trailing [cr|lf]*lf
        if Last > 0 and then Line(Last) = Ascii.Lf then
          for I in reverse Line'First .. Last loop
            if Line(I) = Ascii.Cr or Line(I) = Ascii.Lf then
              Last := I - 1;
            else
              exit;
            end if;
          end loop;
        end if;
        if Debug_Flag then
          Note(":" & Line(Line'First .. Last));
        end if;
        Si := Line'First;
      end if;
      if Si <= Last then
        case State is
          when In_Comment =>
            if Line(Si) = '>'
            and then Si >= Line'First + 2
            and then Line(Si - 2 .. Si) = "-->" then
              State := In_Content;
            end if;
          when In_Content =>
            if Line(Si) = '<' then
              if Si <= Last - 3 and then Line(Si .. Si + 3) = "<!--" then
                State := In_Comment;
                Si := Si + 3;
              elsif Si <= Last - 6
              and then To_Lower(Line(Si .. Si + 6)) = "<script" then
                State := In_Script;
                Si := Si + 6;
              else
                State := In_Bracket;
                Handling_Base := False;
              end if;
            elsif Line(Si) = '"' then
              State := In_Content_String;
            end if;
          when In_Content_String =>
            if Line(Si) = '"' then
              State := In_Content;
            end if;
          when In_Bracket =>
            case To_Lower(Line(Si)) is
              when 'a' .. 'z' =>
                Wi := Si;
                State := In_Bracket_Word;
                Meta_Seen := False;
                Http_Equiv_Seen := False;
              when '"' =>
                State := In_Bracket_String;
              when '>' =>
                State := In_Content;
              when others =>
                null;
            end case;
          when In_Bracket_Word =>
            case To_Lower(Line(Si)) is
              when 'a' .. 'z' =>
                null;
              when others =>
                State := In_Bracket;
                if Si - Wi = 3
                and then To_Lower(Line(Wi .. Si - 1)) = "src" then
                  State := Expect_Eq;
                  Si := Si - 1;
                elsif Si - Wi = 4
                and then To_Lower(Line(Wi .. Si - 1)) = "href" then
                  State := Expect_Eq;
                  Si := Si - 1;
                elsif Si - Wi = 4
                and then To_Lower(Line(Wi .. Si - 1)) = "base" then
                  Handling_Base := True;
                  Si := Si - 1;
                elsif Si - Wi = 4
                and then To_Lower(Line(Wi .. Si - 1)) = "meta" then
                  Meta_Seen := True;
                  Si := Si - 1;
                elsif Si - Wi = 4
                and then To_Lower(Line(Wi .. Si - 1)) = "http"
                and then Si + 5 <= Last
                and then To_Lower(Line(Si .. Si + 5)) = "-equiv" then
                  Http_Equiv_Seen := True;
                  Si := Si + 5; -- 'equiV'
                elsif Meta_Seen and then Http_Equiv_Seen
                and then Si - Wi = 7
                and then To_Lower(Line(Wi .. Si - 1)) = "content" then
                  -- CONTENT="1; URL=http...."
                  for I in Si .. Last - 2 loop
                    if To_Lower(Line(I .. I + 2)) = "url" then
                      State := Expect_Eq;
                      Si := I + 3;
                      exit;
                    elsif Line(I) /= '=' and Line(I) /= '"' and Line(I) /= ';'
                      and Line(I) /= ' ' and Line(I) /= Ascii.Ht
                      and not (Line(I) in '0' .. '9') then
                      exit;
                    end if;
                  end loop;
                  Si := Si - 1;
                else
                  -- unrecognized word
                  Si := Si - 1;
                end if;
            end case;
          when In_Bracket_String =>
            if Line(Si) = '"' then
              State := In_Bracket;
            end if;
          when Expect_Eq =>
            if Line(Si) = '=' then
              State := After_Eq;
            elsif Line(Si) /= ' ' and Line(Si) /= Ascii.Ht then
              State := Skip_To_End_Bracket;
            end if;
            --if state /= expect_eq then
            --Note(line(si) & scanning_states'image(state));
            --end if;
          when After_Eq =>
            if Line(Si) = '"' then
              State := In_Url;
              Di := Target'First;
            elsif Ada.Characters.Handling.Is_Letter(Line(Si)) then
              State := In_Unquoted_Url;
              Di := Target'First;
              Si := Si - 1; -- rescan the letter
            elsif Line(Si) /= ' ' and Line(Si) /= Ascii.Ht then
              State := Skip_To_End_Bracket;
            end if;
          when In_Url | In_Unquoted_Url =>
            if Line(Si) = '"' or Line(Si) = '>' or Line(Si) = ' ' then
              if Line(Si) = '"' then
                if State = In_Unquoted_Url then
                  -- allow href=mailto:"nn">
                  if Si > Line'First and then Line(Si - 1) = ':' then
                    State := In_Url;
                  else
                    Note("Quote terminated! " & Line(Line'First .. Last));
                  end if;
                end if;
                State := Skip_To_End_Bracket;
              elsif Line(Si) = '>' then
                -- even if in quotes!
                if State = In_Url then
                  Note("> terminated! " & Line(Line'First .. Last));
                end if;
                State := In_Content;
              elsif State = In_Unquoted_Url and Line(Si) = ' ' then
                State := Skip_To_End_Bracket;
              end if;
              -- Trim leading blanks from URL: (Otherwise, we keep appending
              -- the Source_Name forever, as we keep getting the same file
              -- each time, but with a longer URL.)
              while Di > 1 and then Target(1) = ' ' loop
                Target(1 .. Di - 2) := Target(2 .. Di - 1);
                Di := Di - 1;
              end loop;
              if Di = 0 then
                null; -- No URL!
              elsif Handling_Base then
                declare
                  Old_Base: Base_Ptr := Current_Base;
                begin
                  Current_Base
                    := new String'(Http_Urls.Scheme(Target(1 .. Di - 1),
                                                    Old_Base.all)
                                   & Http_Urls.Host(Target(1 .. Di - 1),
                                                    Old_Base.all)
                                   & Http_Urls.Path(Target(1 .. Di - 1),
                                                    Old_Base.all));
                  Free(Old_Base);
                exception
                  when others =>
                    Free(Old_Base);
                    raise;
                end;
              else
                Put_Ref(Target(1 .. Di - 1), Current_Base.all);
              end if;
            else
              Target(Di) := Line(Si);
              Di := Di + 1;
            end if;
          when Skip_To_End_Bracket =>
            if Line(Si) = '>' then
              State := In_Content;
            end if;
          when In_Script =>
            if Line(Si) = '<' then
              if Si <= Last - 7
              and then To_Lower(Line(Si .. Si + 7)) = "</script" then
                State := Skip_To_End_Bracket;
                Si := Si + 7;
                -- else probably script code, ignore it.
              end if;
            elsif Si <= Last - 1 and then Line(Si .. Si + 1) = "//" then
              -- A script comment, skip this line.
              Si := Last + 1;
            elsif Si <= Last - 14
            and then To_Lower(Line(Si .. Si + 14)) = "topmenu.makesub" then
              State := In_Script_Menu;
              Si := Si + 14;
              --Note(": found menu in " & Line(Line'first .. Last));
            elsif Si <= Last - 15
            and then To_Lower(Line(Si .. Si + 15)) = "topmenu.makemain" then
              State := In_Script_Menu;
              Si := Si + 15;
              --Note(": found menu in " & Line(Line'first .. Last));
            end if;
          when In_Script_Menu =>
            -- The links are in calls to Topmenu.makeSub and Topmenu.makeSubSub.
            --   TopMenu.makeSubSub(0,0,2,'Helicoptors','atwork/helicop.htm',6)
            -- We scan for the second single quoted string:
            if Line(Si) = '(' then
              State := In_Script_Menu_Args;
              --Note(": found menu args");
            elsif Line(Si) = ')' then
              State := In_Script;
              --Note(": menu args done (1)");
              -- else skip everything else.
            end if;
          when In_Script_Menu_Args =>
            if Line(Si) = ''' then
              -- First quoted string. Discard it.
              Si := Si + 1;
              while Si <= Last loop
                exit when Line(Si) = ''';
                Si := Si + 1;
              end loop;
              State := In_Script_Menu_Args_Qs;
              --Note(": found menu args first quoted string");
            elsif Line(Si) = ')' then
              -- End of arguments, we're done.
              State := In_Script;
              --Note(": menu args done (2)");
              -- else skip everything else.
            end if;
          when In_Script_Menu_Args_Qs =>
            if Line(Si) = ''' then
              -- Second quoted string: starts a link!
              Di := 1;
              State := In_Script_Menu_Args_Link;
              --Note(": found menu args second quoted string");
            elsif Line(Si) = ')' then
              -- End of arguments, we're done.
              State := In_Script;
              --Note(": menu args done (3)");
              -- else skip everything else.
            end if;
          when In_Script_Menu_Args_Link =>
            if Line(Si) = ''' then
              -- Possible end of second quoted string!
              if Si /= 1 and then Line(Si - 1) = '\' then
                -- Quoted quote. Replace the '\' with the quote in Target.
                Target(Di - 1) := ''';
              else
                -- End of quoted string!
                State := In_Script_Menu_Args;
                if Note_Found_Menu_Args_Link then
                  Note(": found menu args link, target="
                       & Target(1 .. Di - 1));
                end if;
                -- Remove leading blanks:
                while Di > 1 and then Target(1) = ' ' loop
                  Target(1 .. Di - 2) := Target(2 .. Di - 1);
                  Di := Di - 1;
                end loop;
                if Di = 0 then
                  null; -- No URL!
                else
                  Put_Ref(Target(1 .. Di - 1), Current_Base.all);
                end if;
              end if;
            else
              Target(Di) := Line(Si);
              Di := Di + 1;
            end if;
        end case;
        Si := Si + 1;
      end if;
    end loop;
    Free(Current_Base);
  exception
    when others =>
      Free(Current_Base);
      raise;
  end Get_Links;

end Html_Scan;
