
package body Str_Sto.Store is

   Next  : Order_Positive := 1;
   Next_ID : IDs := IDs'first;
   IDs_to_Indices : array(IDs) of String_Indices;
   Order : Ordered_Lists(1 .. IDs_to_Indices'length);
   Cache_Last_Present : Order_Natural := 0;
   Cache_Last_Length : Order_Natural := 0;

  function Count return String_Counts is
  -- return the number of distinct strings in this instantiation.
  begin return String_Counts(Next-1);end Count;

  function Is_Present(S : in String) return Boolean is
    Min : Order_Positive := 1;
    Max : Order_Natural := Next-1;  -- zero for empty table
    Guess: Order_Positive;
  begin
    while Min <= Max loop
      Guess := Min+(Max-Min)/2;
      case Str_Sto.Compare(IDs_to_Indices(Order(Guess)), S) is
        when LT => Min := Guess+1;
        when GT => Max := Guess-1;
        when EQ =>
          Cache_Last_Present := Guess;
          Cache_Last_Length := S'length;
          return True;
      end case;
    end loop;
    Cache_Last_Present := 0;
    return False;
  end Is_Present;

  function Put(S : String) return IDs is
  -- Add to store if it isn't there yet.  Return new or old ID.
  -- Raise Overflow if too many or too long strings.
    Min : Order_Positive := 1;
    Max : Order_Natural := Next-1;  -- zero for empty table
    Guess: Order_Positive;
    Index : Str_Sto.String_Indices;
  begin
    if Cache_Last_Present /= 0
    and then Cache_Last_Length = S'length
    and then Str_Sto.Compare(IDs_to_Indices(Order(Cache_Last_Present)), S) = EQ then
      Guess := Cache_Last_Present;
      Cache_Last_Present := 0;
      return Order(Guess);
    else
      Cache_Last_Present := 0;
    end if;
    while Min <= Max loop
      Guess := Min+(Max-Min)/2;
      case Str_Sto.Compare(IDs_to_Indices(Order(Guess)), S) is
        when LT => Min := Guess+1;
        when GT => Max := Guess-1;
        when EQ =>
          return Order(Guess);
      end case;
    end loop;
    if Next >= Order'last then
      raise Str_Sto.Overflow;
    end if;
    Index := Str_Sto.Add(S);
    Order(Min+1 .. Next) := Order(Min .. Next-1);
    Order(Min) := Next_ID;
    IDs_to_Indices(Next_ID) := Index;
    Next := Next+1;
    Next_ID := IDs'succ(Next_ID);
    return IDs'pred(Next_ID);
  end Put;

  function Get(ID : IDs) return String is
  -- raise Nonesuch if no such ID.
  begin
    if ID in Next_ID .. IDs'last then  -- we ain't there yet!
      raise Str_Sto.Nonesuch;
    end if;
    return Str_Sto.Get(IDs_to_Indices(ID));
  end Get;

  function Get_Length(ID : IDs) return Natural is
  -- raise Nonesuch if no such ID.
  begin
    if ID in Next_ID .. IDs'last then  -- we ain't there yet!
      raise Str_Sto.Nonesuch;
    end if;
    return Str_Sto.Get_Length(IDs_to_Indices(ID));
  end Get_Length;

  function "<"(Left, Right : IDs) return Boolean is
  begin
    return (Str_Sto.Compare(IDs_to_Indices(Left), IDs_to_Indices(Right)) = LT);
  end "<";

  function Get_Ordered_Contents return Ordered_Lists is
  -- returns array of all IDs in table, in order.
  begin return Order(1 .. Next-1);end Get_Ordered_Contents;

end Str_Sto.Store;
