with Ada.Exceptions,
     Claw.Sockets;
package Text_Line_Sockets is
  type Socket_Type(Max_Line_Length: Positive)
  is new Claw.Sockets.Socket_Type with private;

  procedure Get_Line(Socket  : in out Socket_Type;
                     Line    :    out String;
                     Last    :    out Natural;
                     Timeout : in     Duration := 30.0);
    -- Return Line(Line'first .. Last)
    -- Input up through LF if LF occurs,
    -- Line(Line'range) if no LF seen,
    -- or Input up through last received char if timeout or hangup
    -- NOTE Timeout parameter is last (unlike Claw.Sockets.Get) for defaulting

  private

  type Statuses is (OK, Problem, EOF);

  type Socket_Type(Max_Line_Length: Positive)
  is new Claw.Sockets.Socket_Type with record
    Buffer : String(1 .. Max_Line_Length);
    Last   : Natural := 0;
    Next   : Positive := 1;
    Status : Statuses := OK;
    Occurrence : Ada.Exceptions.Exception_Occurrence;
  end record;

  procedure Close(Socket : in out Socket_Type);

end Text_Line_Sockets;
